/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;
import java.util.List;
import java.util.ArrayList;

/**
<pre>{@code
                   y
                   |
                   |
                   | v[4]
                 1 +-----------------+ v[5]=(1,1,0)
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
      v[7] +-----------------+ v[6]  |
           |       |         |       |                    y
           |       |         |       |                    |
           |       |         |       | v[1]               |
           |  v[0] +---------|-------+------> x           |
           |      /          |      /1                    |
           |    /            |    /                       +-----> x
           |  /              |  /                        /
           |/                |/                         /
         1 +-----------------+                         /
         / v[3]=(0,0,1)      v[2]=(1,0,1)            z
       /
     /
    z
}</pre>
   Render a single wireframe cube.
<p>
   What is wrong with this version of the program?
   Why does the cube become distorted when you move
   it around?
*/
@SuppressWarnings("serial")
public class InteractiveCube_v3_R2 extends InteractiveFrame
{
   private Scene scene;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
   */
   public InteractiveCube_v3_R2(String title, int fbWidth, int fbHeight)
   {
      super(title, fbWidth, fbHeight);

      // Create a Model object to hold the geometry.
      Model model = new Model();

      // Create the geometry for the Model.
      Vertex v0, v1, v2, v3, v4, v5, v6, v7;
      v0 = new Vertex(0.0, 0.0, -2.0);  // four vertices around the bottom face
      v1 = new Vertex(1.0, 0.0, -2.0);
      v2 = new Vertex(1.0, 0.0, -1.0);
      v3 = new Vertex(0.0, 0.0, -1.0);
      v4 = new Vertex(0.0, 1.0, -2.0);  // four vertices around the top face
      v5 = new Vertex(1.0, 1.0, -2.0);
      v6 = new Vertex(1.0, 1.0, -1.0);
      v7 = new Vertex(0.0, 1.0, -1.0);

      // Add the geometry to the Model.
      List<Vertex> vertexList = new ArrayList<>();
      vertexList.add(v0);
      vertexList.add(v1);
      vertexList.add(v2);
      vertexList.add(v3);
      vertexList.add(v4);
      vertexList.add(v5);
      vertexList.add(v6);
      vertexList.add(v7);
      vertexList.add(v2);
      vertexList.add(v6);

      List<LineSegment> lineSegmentList = new ArrayList<>();
      lineSegmentList.add(new LineSegment(0, 1, model));
      lineSegmentList.add(new LineSegment(1, 2, model));
      lineSegmentList.add(new LineSegment(2, 3, model));
      lineSegmentList.add(new LineSegment(3, 0, model));
      lineSegmentList.add(new LineSegment(4, 5, model));
      lineSegmentList.add(new LineSegment(5, 6, model));
      lineSegmentList.add(new LineSegment(6, 7, model));
      lineSegmentList.add(new LineSegment(7, 4, model));
      lineSegmentList.add(new LineSegment(0, 4, model));
      lineSegmentList.add(new LineSegment(1, 5, model));
      lineSegmentList.add(new LineSegment(3, 7, model));
      lineSegmentList.add(new LineSegment(8, 9, model));

      model.vertexList = vertexList;
      model.lineSegmentList = lineSegmentList;


      // Create the Scene object that we shall render.
      scene = new Scene();

      // Add the Model to the Scene.
      scene.addModel(model);
   }


   // Implement part of the KeyListener interface.
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
       //Rasterize.debug = ! Rasterize.debug;
      }
      else if ('c' == c)
      {
         Pipeline.doClipping = ! Pipeline.doClipping;
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
      }
      else if ('s' == c) // Scale the cube 10% smaller.
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x /= 1.1;  // This does NOT work with
            v.y /= 1.1;  // perspective projection.
            v.z /= 1.1;
         }
      }
      else if ('S' == c) // Scale the cube 10% larger.
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x *= 1.1;  // This does NOT work with
            v.y *= 1.1;  // perspective projection.
            v.z *= 1.1;
         }
      }
      else if ('x' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x -= 0.1;
         }
      }
      else if ('X' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x += 0.1;
         }
      }
      else if ('y' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.y -= 0.1;
         }
      }
      else if ('Y' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.y += 0.1;
         }
      }
      else if ('z' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.z -= 0.1;
         }
      }
      else if ('Z' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.z += 0.1;
         }
      }

      // Render again.
      FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB(Color.black);
      Pipeline.render(scene, fb);
      fbp.update();
      repaint();
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      int w = this.fbp.getWidth();
      int h = this.fbp.getHeight();

      // Create a new FrameBuffer that fits the new window size.
      FrameBuffer fb = new FrameBuffer(w, h);
      this.fbp.setFrameBuffer(fb);
      fb.clearFB(Color.black);
      Pipeline.render(scene, fb);
      fbp.update();
      repaint();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // Define initial dimensions for a FrameBuffer.
      int width  = 512;
      int height = 512;
      // Create an InteractiveFrame containing a FrameBuffer
      // with the given dimensions. NOTE: We need to call the
      // InteractiveCube_v3_R2 constructor in the Java GUI Event
      // Dispatch Thread, otherwise we get a race condition
      // between the constructor (running in the main() thread)
      // and the very first ComponentEvent (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         new Runnable() // an anonymous inner class constructor
         {
            public void run() // implement the Runnable interface
            {
               // call the constructor that builds the gui
               new InteractiveCube_v3_R2("Renderer 2", width, height);
            }
         }
      );
   }//main()


   private static void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the cube along the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the cube.");
      System.out.println("Use the 'c' key to toggle line clipping on and off.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
